from .ard_regression import ARDRegression
from .ada_boost_classifier import AdaBoostClassifier
from .ada_boost_regressor import AdaBoostRegressor
from .bayesian_ridge import BayesianRidge
from .bernoulli_nb import BernoulliNB
from .cca import CCA
from .calibrated_classifier_cv import CalibratedClassifierCV
from .complement_nb import ComplementNB
from .decision_tree_classifier import DecisionTreeClassifier
from .decision_tree_regressor import DecisionTreeRegressor
from .elastic_net_cv import ElasticNetCV
from .elastic_net import ElasticNet
from .extra_trees_classifier import ExtraTreesClassifier
from .extra_trees_regressor import ExtraTreesRegressor
from .gaussian_nb import GaussianNB
from .gaussian_process_classifier import GaussianProcessClassifier
from .gaussian_process_regressor import GaussianProcessRegressor
from .gradient_boosting_classifier import GradientBoostingClassifier
from .gradient_boosting_regressor import GradientBoostingRegressor
from .huber_regressor import HuberRegressor
from .k_neighbors_classifier import KNeighborsClassifier
from .k_neighbors_regressor import KNeighborsRegressor
from .kernel_ridge import KernelRidge
from .label_propagation import LabelPropagation
from .label_spreading import LabelSpreading
from .lars_cv import LarsCV
from .lars import Lars
from .lasso_cv import LassoCV
from .lasso import Lasso
from .lasso_lars_cv import LassoLarsCV
from .lasso_lars import LassoLars
from .lasso_lars_ic import LassoLarsIC
from .linear_discriminant_analysis import LinearDiscriminantAnalysis
from .linear_regression import LinearRegression
from .linear_svc import LinearSVC
from .linear_svr import LinearSVR
from .logistic_regression_cv import LogisticRegressionCV
from .logistic_regression import LogisticRegression
from .mlp_classifier import MLPClassifier
from .mlp_regressor import MLPRegressor
from .multi_task_elastic_net_cv import MultiTaskElasticNetCV
from .multi_task_elastic_net import MultiTaskElasticNet
from .multi_task_lasso_cv import MultiTaskLassoCV
from .multi_task_lasso import MultiTaskLasso
from .multinomial_nb import MultinomialNB
from .nearest_centroid import NearestCentroid
from .nu_svc import NuSVC
from .nu_svr import NuSVR
from .orthogonal_matching_pursuit_cv import OrthogonalMatchingPursuitCV
from .orthogonal_matching_pursuit import OrthogonalMatchingPursuit
from .pls_canonical import PLSCanonical
from .pls_regression import PLSRegression
from .passive_aggressive_classifier import PassiveAggressiveClassifier
from .passive_aggressive_regressor import PassiveAggressiveRegressor
from .perceptron import Perceptron
from .quadratic_discriminant_analysis import QuadraticDiscriminantAnalysis
from .ransac_regressor import RANSACRegressor
from .radius_neighbors_classifier import RadiusNeighborsClassifier
from .radius_neighbors_regressor import RadiusNeighborsRegressor
from .random_forest_classifier import RandomForestClassifier
from .random_forest_regressor import RandomForestRegressor
from .ridge_cv import RidgeCV
from .ridge_classifier_cv import RidgeClassifierCV
from .ridge_classifier import RidgeClassifier
from .ridge import Ridge
from .sgd_classifier import SGDClassifier
from .sgd_regressor import SGDRegressor
from .svc import SVC
from .svr import SVR
from .theil_sen_regressor import TheilSenRegressor
from .transformed_target_regressor import TransformedTargetRegressor
from .additive_chi2_sampler import AdditiveChi2Sampler
from .bernoulli_rbm import BernoulliRBM
from .binarizer import Binarizer
from .birch import Birch
from .dictionary_learning import DictionaryLearning
from .factor_analysis import FactorAnalysis
from .fast_ica import FastICA
from .feature_agglomeration import FeatureAgglomeration
from .function_transformer import FunctionTransformer
from .gaussian_random_projection import GaussianRandomProjection
from .incremental_pca import IncrementalPCA
from .isomap import Isomap
from .k_bins_discretizer import KBinsDiscretizer
from .k_means import KMeans
from .kernel_pca import KernelPCA
from .label_binarizer import LabelBinarizer
from .label_encoder import LabelEncoder
from .latent_dirichlet_allocation import LatentDirichletAllocation
from .locally_linear_embedding import LocallyLinearEmbedding
from .max_abs_scaler import MaxAbsScaler
from .min_max_scaler import MinMaxScaler
from .mini_batch_dictionary_learning import MiniBatchDictionaryLearning
from .mini_batch_k_means import MiniBatchKMeans
from .mini_batch_sparse_pca import MiniBatchSparsePCA
from .missing_indicator import MissingIndicator
from .multi_label_binarizer import MultiLabelBinarizer
from .nmf import NMF
from .normalizer import Normalizer
from .nystroem import Nystroem
from .one_hot_encoder import OneHotEncoder
from .ordinal_encoder import OrdinalEncoder
from .pca import PCA
from .plssvd import PLSSVD
from .polynomial_features import PolynomialFeatures
from .power_transformer import PowerTransformer
from .quantile_transformer import QuantileTransformer
from .rbf_sampler import RBFSampler
from .robust_scaler import RobustScaler
from .simple_imputer import SimpleImputer
from .skewed_chi2_sampler import SkewedChi2Sampler
from .sparse_pca import SparsePCA
from .sparse_random_projection import SparseRandomProjection
from .standard_scaler import StandardScaler
from .truncated_svd import TruncatedSVD



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.operators
import lale.docstrings
import sklearn.kernel_approximation

class NystroemImpl():
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.kernel_approximation.Nystroem(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema for the Nystroem model from scikit-learn.',
    'allOf': [{
        'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'additionalProperties': False,
        'required': ['kernel', 'gamma', 'coef0', 'degree', 'n_components', 'random_state'],
        'relevantToOptimizer': ['kernel', 'gamma', 'coef0', 'degree', 'n_components'],
        'properties': {
            'kernel': {
                'description': 'Kernel map to be approximated. In the scikit learn version, this can be a string or a callable. To keep arguments as plain JSON documents, the wrapper only allows an enum of the keys of sklearn.metrics.pairwise.KERNEL_PARAMS.',
                'enum': ['additive_chi2', 'chi2', 'cosine', 'linear', 'poly', 'polynomial', 'rbf', 'laplacian', 'sigmoid'],
                'default': 'rbf'},
            'gamma': {
                'description': 'Gamma parameter.',
                'anyOf': [{
                    'enum': [None]}, {
                    'type': 'number',
                    'distribution': 'loguniform',
                    'minimumForOptimizer': 3.0517578125e-05,
                    'maximumForOptimizer': 8}],
                'default': None},
            'coef0': {
                'description': 'Zero coefficient.',
                'anyOf': [{
                    'enum': [None]}, {
                    'type': 'number',
                    'minimum': (- 1),
                    'distribution': 'uniform',
                    'maximumForOptimizer': 1}],
                'default': None},
            'degree': {
                'description': 'Degree of the polynomial kernel.',
                'anyOf': [{
                    'enum': [None]}, {
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 5}],
                'default': None},
            'kernel_params':{
                'description': 'Additional parameters (keyword arguments) for kernel '
                'function passed as callable object.',
                'anyOf':[
                {'type':'object'},
                {'enum':[None]}],
                'default': None
            },
            'n_components': {
                'description': 'Number of features to construct. How many data points will be used to construct the mapping.',
                'type': 'integer',
                'default': 100,
                'minimum': 1,
                'distribution': 'uniform',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 256},
            'random_state': {
                'description': 'Seed of pseudo-random number generator.',
                'anyOf': [{
                    'description': 'RandomState used by np.random',
                    'enum': [None]}, {
                    'description': 'Explicit seed.',
                    'type': 'integer'}],
                'default': None},
        }},
    {   'description': 'Gamma is ignored by other kernels.',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'gamma': {
                    'enum': [None]},
            }}, {
            'type': 'object',
            'properties': {
                'kernel': {
                    'enum': ['rbf', 'laplacian', 'polynomial', 'additive_chi2', 'sigmoid']},
            }}]},
    {   'description': 'Zero coefficient ignored by other kernels.',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'coef0': {
                    'enum': [None]},
            }}, {
            'type': 'object',
            'properties': {
                'kernel': {
                    'enum': ['polynomial', 'sigmoid']},
            }}]},
    {   'description': 'Degree ignored by other kernels.',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'degree': {
                    'enum': [None]},
            }}, {
            'type': 'object',
            'properties': {
                'kernel': {
                    'enum': ['polynomial']},
            }}]}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training the Nystroem model from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {'type': 'number'}}},
        'y': {
            'description': 'Target class labels; the array is over samples.'}}}

_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for predictions using the Nystroem model from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {'type': 'number'}}}}}

_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for predictions (projected data) using the Nystroem model from scikit-learn.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Nystroem`_ transformer from scikit-learn.

.. _`Nystroem`: https://scikit-learn.org/0.20/modules/generated/sklearn.kernel_approximation.Nystroem.html#sklearn-kernel-approximation-nystroem
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.nystroem.html',
    'type': 'object',
    'tags': {
        'pre': ['~categoricals'],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(NystroemImpl, _combined_schemas)

Nystroem = lale.operators.make_operator(NystroemImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.cluster.hierarchical
import lale.docstrings
import lale.operators
import numpy as np

class FeatureAgglomerationImpl():

    def __init__(self, n_clusters=2, affinity='euclidean', memory=None, connectivity=None, compute_full_tree=None, linkage='ward', pooling_func=None):
        self._hyperparams = {
            'n_clusters': n_clusters,
            'affinity': affinity,
            'memory': memory,
            'connectivity': connectivity,
            'compute_full_tree': compute_full_tree,
            'linkage': linkage,
            'pooling_func': pooling_func}
        self._wrapped_model = sklearn.cluster.hierarchical.FeatureAgglomeration(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Agglomerate features.',
    'allOf': [{
        'type': 'object',
        'required': ['memory', 'compute_full_tree', 'pooling_func'],
        'relevantToOptimizer': ['n_clusters', 'affinity', 'compute_full_tree', 'linkage'],
        'additionalProperties': False,
        'properties': {
            'n_clusters': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 8,
                'default': 2,
                'description': 'The number of clusters to find.'},
            'affinity': {
                'anyOf': [{
                    'enum': ['euclidean', 'l1', 'l2', 'manhattan', 'cosine',
                    'precomputed']}, {
                    'forOptimizer':False,
                    'type': 'object' }],#callable
                'default': 'euclidean',
                'description': 'Metric used to compute the linkage. Can be "euclidean", "l1", "l2",'},
            'memory': {
                'anyOf': [{
                    'type': 'string'}, {
                    'forOptimizer':False,
                    'type': 'object' }, { #object with the joblib.Memory interface
                    'enum':[None]}],  
                'default': None,             
                'description': 'Used to cache the output of the computation of the tree.'},
            'connectivity': {
                'anyOf': [{
                    'type': 'array',
                    'items': { 'type': 'array', 'items': { 'type': 'number' }}}, {
                    'forOptimizer':False,
                    'type': 'object' #a callable that transforms the data into a connectivity matrix, 
                                     #such as derived from kneighbors_graph
                    }, {'enum': [None]}],
                'default': None,
                'description': 'Connectivity matrix. Defines for each feature the neighboring'},
            'compute_full_tree': {
                'anyOf':[{
                    'type': 'boolean'
                },{
                    'enum':['auto']
                }],
                'default': 'auto',
                'description': 'Stop early the construction of the tree at n_clusters. This is'},
            'linkage': {
                'enum': ['ward', 'complete', 'average', 'single'],
                'default': 'ward',
                'description': 'Which linkage criterion to use. The linkage criterion determines which'},
            'pooling_func': {
                'description': 'This combines the values of agglomerated features into a single',
                'default': np.mean},
        }}, {
        'description': 'affinity, if linkage is "ward", only "euclidean" is accepted',
          'anyOf': [
            { 'type': 'object',
              'properties': {'affinity': {'enum': ['euclidean']}}},
            { 'type': 'object',
              'properties': {
                'linkage':{'not': {'enum': ['ward']}}}}]},{
        'description': 'compute_full_tree, useful only when specifying a connectivity matrix',
        'anyOf': [
        { 'type': 'object',
            'properties': {'compute_full_tree': {'not': {'enum': ['True']}}}},
        { 'type': 'object',
            'properties': {
            'connectivity': {'not': {'enum': ['None']}}}}]        
        },
        {'description': 'n_clusters must be None if distance_threshold is not None.',
        'anyOf': [
        { 'type': 'object',
            'properties': {'n_clusters': {'enum': ['None']}}},
        { 'type': 'object',
            'properties': {
            'distance_threshold': {'enum': ['None']}}}]
        },
        {'description': 'compute_full_tree must be True if distance_threshold is not None.',
        'anyOf': [
        { 'type': 'object',
            'properties': {'compute_full_tree': {'enum': ['True']}}},
        { 'type': 'object',
            'properties': {
            'distance_threshold': {'enum': ['None']}}}]
        }],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the hierarchical clustering on the data',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data'},
        'y': {'description': 'Ignored'},
}}

_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform a new matrix using the built clustering',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'A M by N array of M observations in N dimensions or a length'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The pooled values for each feature cluster.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Feature agglomeration`_ transformer from scikit-learn.

.. _`Feature agglomeration`: https://scikit-learn.org/0.20/modules/generated/sklearn.cluster.FeatureAgglomeration.html#sklearn-cluster-featureagglomeration
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.feature_agglomeration.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(FeatureAgglomerationImpl, _combined_schemas)

FeatureAgglomeration = lale.operators.make_operator(FeatureAgglomerationImpl, _combined_schemas)




from sklearn.preprocessing._discretization import KBinsDiscretizer as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class KBinsDiscretizerImpl():

    def __init__(self, n_bins=5, encode='onehot', strategy='quantile'):
        self._hyperparams = {
            'n_bins': n_bins,
            'encode': encode,
            'strategy': strategy}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for KBinsDiscretizer    Bin continuous data into intervals.',
    'allOf': [{
        'type': 'object',
        'required': ['n_bins', 'encode', 'strategy'],
        'relevantToOptimizer': [],
        'additionalProperties': False,
        'properties': {
            'n_bins': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }],
                'default': 5,
                'description': 'The number of bins to produce'},
            'encode': {
                'enum': ['onehot', 'onehot-dense', 'ordinal'],
                'default': 'onehot',
                'description': 'Method used to encode the transformed result'},
            'strategy': {
                'enum': ['uniform', 'quantile', 'kmeans'],
                'default': 'quantile',
                'description': 'Strategy used to define the widths of the bins'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fits the estimator.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'laleType': 'Any',
            'XXX TODO XXX': 'numeric array-like, shape (n_samples, n_features)',
            'description': 'Data to be discretized.'},
        'y': {
            'laleType': 'Any',
            'XXX TODO XXX': 'ignored'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Discretizes the data.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'laleType': 'Any',
            'XXX TODO XXX': 'numeric array-like, shape (n_samples, n_features)',
            'description': 'Data to be discretized.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Data in the binned space.',
    'laleType': 'Any',
    'XXX TODO XXX': 'numeric array-like or sparse matrix',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.KBinsDiscretizer#sklearn-preprocessing-kbinsdiscretizer',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(KBinsDiscretizerImpl, _combined_schemas)
KBinsDiscretizer = lale.operators.make_operator(KBinsDiscretizerImpl, _combined_schemas)




# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import autoai_libs.cognito.transforms.transform_utils
import lale.docstrings
import lale.operators

class TB1Impl():
    def __init__(self, tans_class, name, datatypes, feat_constraints, tgraph, apply_all, col_names, col_dtypes, col_as_json_objects):
        self._hyperparams = {
            'tans_class': tans_class,
            'name': name,
            'datatypes': datatypes,
            'feat_constraints': feat_constraints,
            'tgraph': tgraph,
            'apply_all': apply_all,
            'col_names': col_names,
            'col_dtypes': col_dtypes,
            'col_as_json_objects': col_as_json_objects}
        self._wrapped_model = autoai_libs.cognito.transforms.transform_utils.TB1(**self._hyperparams)

    def fit(self, X, y=None, **fit_params):
        if fit_params is None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X, y, **fit_params)
        return self

    def transform(self, X):
        result = self._wrapped_model.transform(X)
        return result

_hyperparams_schema = {
    'allOf': [{
        'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'additionalProperties': False,
        'required': ['tans_class', 'name', 'datatypes', 'feat_constraints', 'tgraph', 'apply_all', 'col_names', 'col_dtypes', 'col_as_json_objects'],
        'relevantToOptimizer': [],
        'properties': {
            'tans_class': {
                'description': 'A class that implements fit() and transform() in accordance with the transformation function definition.',
                'laleType': 'Any',
                'default': None},
            'name': {
                'description': 'A string name that uniquely identifies this transformer from others.',
                'anyOf': [
                {   'type': 'string'},
                {   'enum': [None]}],
                'default': None},
            'datatypes': {
                'description': 'List of datatypes that are valid input to the transformer function (numeric, float, int, etc.).',
                'anyOf': [
                {   'type': 'array', 'items': {'type': 'string'}},
                {   'enum': [None]}],
                'default': None},
            'feat_constraints': {
                'description': 'All constraints that must be satisfied by a column to be considered a valid input to this transform.',
                'laleType': 'Any',
                'default': None},
            'tgraph': {
                'description': 'Should be the invoking TGraph() object.',
                'anyOf': [
                {   'laleType': 'Any'},
                {   'enum': [None],
                    'description': 'Passing None will result in some failure to detect some inefficiencies due to lack of caching.'}],
                'default': None},
            'apply_all': {
                'description': 'Only use applyAll = True. It means that the transformer will enumerate all features (or feature sets) that match the specified criteria and apply the provided function to each.',
                'type': 'boolean',
                'default': True},
            'col_names': {
                'description': 'Names of the feature columns in a list.',
                'anyOf': [
                {   'type': 'array', 'items': {'type': 'string'}},
                {   'enum': [None]}],
                'default': None},
            'col_dtypes': {
                'description': 'List of the datatypes of the feature columns.',
                'anyOf': [
                {   'type': 'array', 'items': {'laleType': 'Any'}},
                {   'enum': [None]}],
                'default': None},
            'col_as_json_objects': {
                'description': 'Names of the feature columns in a json dict.',
                'anyOf': [
                {   'type': 'object'},
                {   'enum': [None]}],
                'default': None}}}]}

_input_fit_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'laleType': 'Any'}}},
        'y': {
            'laleType': 'Any'}}}

_input_transform_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'laleType': 'Any'}}}}}

_output_transform_schema = {
    'description': 'Features; the outer array is over samples.',
    'type': 'array',
    'items': {'type': 'array', 'items': {'laleType': 'Any'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """Operator from `autoai_libs`_. Feature transformation for unary state-based transformations (with fit/transform), such as frequent count.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.tb1.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(TB1Impl, _combined_schemas)

TB1 = lale.operators.make_operator(TB1Impl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Generic, Iterator, Iterable, List, Optional, Sequence, Tuple, TypeVar, Union, overload
import json
import jsonschema
import random
import numpy as np
from enum import Enum

Freqs = Dict[str, int]
PGO = Dict[str, Dict[str, Freqs]]

class DefaultValue(Enum):
    token = 0
_default_value = DefaultValue.token

Def = TypeVar('Def')
Defaultable = Union[DefaultValue, Def]


def remove_defaults(it):
    return filter((lambda x: x[1] is not _default_value), it)

XDK = TypeVar('XDK')
XDV = TypeVar('XDV')
def remove_defaults_dict(d:Dict[XDK, Defaultable[XDV]])->Dict[XDK, XDV]:
    return dict(remove_defaults(d.items()))

# utilites to load a pgo from json-ish
def load_pgo_file(filepath) -> PGO:
    with open(filepath) as json_file:
        json_data = json.load(json_file)
        return load_pgo_data(json_data)

def load_pgo_data(json_data) -> PGO:
    jsonschema.validate(json_data, _input_schema, jsonschema.Draft4Validator)
    norm = normalize_pgo_type(json_data)
    return norm

# TODO: Add support for falling back on an underlying distribution
# with some probability
T = TypeVar('T')
class FrequencyDistribution(Generic[T]):
    """ Represents the distribution implied by a histogram
    """
    freq_dist:np.array # Array[T,int]
    vals:np.array # Array[T]
    cumulative_freqs:np.array # Array[int] 

    @classmethod
    def asIntegerValues(cls, freqs:Iterable[Tuple[Any, int]], inclusive_min:Optional[int]=None, inclusive_max:Optional[int]=None)->'FrequencyDistribution[int]':
        freqs = freqsAsIntegerValues(freqs, inclusive_min=inclusive_min, inclusive_max=inclusive_max)
        return FrequencyDistribution[int](list(freqs), dtype=int)

    @classmethod
    def asFloatValues(cls, freqs:Iterable[Tuple[Any, int]], inclusive_min:Optional[int]=None, inclusive_max:Optional[int]=None)->'FrequencyDistribution[float]':
        freqs = freqsAsFloatValues(freqs, inclusive_min=inclusive_min, inclusive_max=inclusive_max)
        return FrequencyDistribution[float](list(freqs), dtype=float)

    @classmethod
    def asEnumValues(cls, freqs:Iterable[Tuple[Any, int]], values:List[Any])->'FrequencyDistribution[Any]':
        freqs = freqsAsEnumValues(freqs, values=values)
        return FrequencyDistribution[Any](list(freqs), dtype=object)

    def __init__(self, freqs:Iterable[Tuple[Defaultable[T], int]], dtype=object):
        # we need them to be sorted for locality
        sorted_freq_list = sorted(freqs, key = (lambda k: (k[0] is _default_value, None if k[0] is _default_value else k[0])))
        freqs_array = np.array(sorted_freq_list, dtype=[('value',object),('frequency',int)])
#        freqs_array.sort(order='value')
        self.freq_dist = freqs_array

        self.vals = freqs_array['value']
        self.cumulative_freqs = np.cumsum(freqs_array['frequency'])

    def __len__(self):
        return np.int_(self.cumulative_freqs[-1])
    
    @overload
    def __getitem__(self, key: int) -> T: ...

    @overload
    def __getitem__(self, key: Sequence[int]) -> Sequence[T]: ...

    @overload
    def __getitem__(self, key: slice) -> Sequence[T]: ...

    def __getitem__(self, key: Union[int, Sequence[int], slice]) -> Union[T, Sequence[T]]:
        indices:Sequence[int]
        single = False
        if isinstance(key, (int, float)):
            single = True
            indices = [key]
        elif isinstance(key, slice):
            # TODO: this could be made more efficient
            indices = range(key.start or 0, key.stop or len(self), key.step or 1)
        else:
            indices = key

        val_indices:Sequence[int] = np.searchsorted(self.cumulative_freqs, indices, side='right')

        values = self.vals[val_indices].tolist()

        if single:
            assert len(values) == 1
            return values[0]
        else:
            return values

    def sample(self)->T:
        l = len(self)
        i = random.randrange(l)
        return self[i]

    def samples(self, count:int)->Sequence[T]:
        l = len(self)
        i:Sequence[int] = [random.randrange(l) for _ in range(count)]
        return self[i]


# utiltities to convert and sample from a PGO frequency distribution

DEFAULT_STR = "default"

def freqsAsIntegerValues(freqs:Iterable[Tuple[Any, int]], inclusive_min:Optional[int]=None, inclusive_max:Optional[int]=None)->Iterator[Tuple[Defaultable[int], int]]:
    """ maps the str values to integers, and skips anything that does not look like an integer"""
    for v,f in freqs:
        try:
            if v == DEFAULT_STR:
                yield _default_value, f
                continue
            i = int(v)
            if inclusive_min is not None and inclusive_min > i:
                continue
            if inclusive_max is not None and inclusive_max < i:
                continue
            yield i,f
        except:
            pass

def freqsAsFloatValues(freqs:Iterable[Tuple[Any, int]], inclusive_min:Optional[float]=None, inclusive_max:Optional[float]=None)->Iterator[Tuple[Defaultable[float], int]]:
    """ maps the str values to integers, and skips anything that does not look like an integer"""
    for v,f in freqs:
        try:
            if v == DEFAULT_STR:
                yield _default_value, f
                continue
            i = float(v)
            if inclusive_min is not None and inclusive_min > i:
                continue
            if inclusive_max is not None and inclusive_max < i:
                continue

            yield i,f
        except:
            pass

# TODO: we can get a dictionary from freqs (before items() was called)
# and then lookup values in it (since values is likely smaller then freqs)
# or, of course, check which one is smaller and iterate through it
def freqsAsEnumValues(freqs:Iterable[Tuple[Any, int]],values:List[Any])->Iterator[Tuple[Defaultable[Any], int]]:
    """ only keeps things that match the string representation of values in the enumeration.
        converts from the string to the value as represented in the enumeration.
    """
    def as_str(v)->str:
        """ There are some quirks in how the PGO files
        encodes values relative to python's str method
        """
        if v is None:
            return "none"
        elif v is True:
            return "true"
        elif v is False:
            return "false"
        else:
            return str(v)
    value_lookup = {as_str(k):k for k in values}
    for v,f in freqs:
        if v == DEFAULT_STR:
            yield _default_value, f
            continue

        if v in value_lookup:
            yield value_lookup[v],f

_input_type = Dict[str, Dict[str, Union[int, Dict[str, Union[str, int]]]]]

# For now, we skip things of the form
# alg -> {default: number}
# (i.e. without parameters)
def normalize_pgo_type(data: _input_type) -> PGO:
    return {alg: {param_keys: {param_values: int(param_counts)
                               for param_values, param_counts in v2.items()}
                  for param_keys, v2 in v1.items() if isinstance(v2, dict)}
            for alg, v1 in data.items()}


_input_schema: Any = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input format for pgo files.  Keys are the name of the algorithm',
    'type': 'object',
    'additionalProperties': {
        "anyOf": [{
            'description': 'Keys are the parameter names',
            'type': 'object',
            'additionalProperties': {
                'description': 'Keys are value names',
                'type': 'object',
                'additionalProperties': {
                    'anyOf': [{
                        'description': 'the number of times this value was found',
                        'type': 'integer'
                    },
                        {
                        'description': 'the number of times this value was found',
                        'type': 'string'
                    }]
                }
            }
        },
            {
            'description': 'default value for the optimizer',
            'type': 'object',
            'additionalProperties': False,
            'required': ['default'],
            'properties': {
                    'default': {'anyOf': [{
                        'description': 'the number of times the default was found',
                        'type': 'integer'
                    },
                        {
                        'description': 'the number of times the default was found',
                        'type': 'string'
                    }]}
            }
        }]
    }
}



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators

class NoOpImpl():
    def __init__(self):
        pass

    def transform(self, X):
        return X

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        return s_X

_hyperparams_schema = {
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints, if any.',
      'type': 'object',
      'additionalProperties': False,
      'relevantToOptimizer': [],
      'properties': {}}]}

_input_transform_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {'description': 'Features; no restrictions on data type.'}}}

_output_transform_schema = {
    'description': 'Features; no restrictions on data type.',
    'laleType': 'Any'}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Passes the data through unchanged.',
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.no_op.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(NoOpImpl, _combined_schemas)

NoOp = lale.operators.make_operator(NoOpImpl, _combined_schemas)




from sklearn.semi_supervised.label_propagation import LabelPropagation as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class LabelPropagationImpl():

    def __init__(self, kernel='rbf', gamma=20, n_neighbors=7, alpha=None, max_iter=1000, tol=0.001, n_jobs=None):
        self._hyperparams = {
            'kernel': kernel,
            'gamma': gamma,
            'n_neighbors': n_neighbors,
            'alpha': alpha,
            'max_iter': max_iter,
            'tol': tol,
            'n_jobs': n_jobs}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for LabelPropagation    Label Propagation classifier',
    'allOf': [{
        'type': 'object',
        'required': ['kernel', 'gamma', 'n_neighbors', 'alpha', 'max_iter', 'tol', 'n_jobs'],
        'relevantToOptimizer': ['kernel', 'gamma', 'n_neighbors', 'max_iter', 'tol'],
        'additionalProperties': False,
        'properties': {
            'kernel': {
                'enum': ['knn', 'rbf', 'callable'],
                'default': 'rbf',
                'description': 'String identifier for kernel function to use or the kernel function itself'},
            'gamma': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 20.0,
                'distribution': 'uniform',
                'default': 20,
                'description': 'Parameter for rbf kernel'},
            'n_neighbors': {
                'XXX TODO XXX': 'integer > 0',
                'description': 'Parameter for knn kernel',
                'type': 'integer',
                'minimumForOptimizer': 5,
                'maximumForOptimizer': 20,
                'distribution': 'uniform',
                'default': 7},
            'alpha': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Clamping factor'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'Change maximum number of iterations allowed'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.001,
                'description': 'Convergence tolerance: threshold to consider the system at steady state'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of parallel jobs to run'},
        }}, {
        'XXX TODO XXX': "Parameter: kernel > only 'rbf' and 'knn' strings are valid inputs"}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit a semi-supervised label propagation model based',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'A {n_samples by n_samples} size matrix will be created from this'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'n_labeled_samples (unlabeled points are marked as -1) All unlabeled samples will be transductively assigned labels'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Performs inductive inference across the model.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predictions for input data',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict probability for each possible outcome.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Normalized probability distributions across class labels',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.semi_supervised.LabelPropagation#sklearn-semi_supervised-labelpropagation',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema},
}
lale.docstrings.set_docstrings(LabelPropagationImpl, _combined_schemas)
LabelPropagation = lale.operators.make_operator(LabelPropagationImpl, _combined_schemas)





from sklearn.linear_model.passive_aggressive import PassiveAggressiveClassifier as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class PassiveAggressiveClassifierImpl():

    def __init__(self, C=1.0, fit_intercept=True, max_iter=None, tol=None, early_stopping=False, validation_fraction=0.1, n_iter_no_change=5, shuffle=True, verbose=0, loss='hinge', n_jobs=None, random_state=None, warm_start=False, class_weight='balanced', average=False, n_iter=None):
        self._hyperparams = {
            'C': C,
            'fit_intercept': fit_intercept,
            'max_iter': max_iter,
            'tol': tol,
            'early_stopping': early_stopping,
            'validation_fraction': validation_fraction,
            'n_iter_no_change': n_iter_no_change,
            'shuffle': shuffle,
            'verbose': verbose,
            'loss': loss,
            'n_jobs': n_jobs,
            'random_state': random_state,
            'warm_start': warm_start,
            'class_weight': class_weight,
            'average': average,
            'n_iter': n_iter}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for PassiveAggressiveClassifier    Passive Aggressive Classifier',
    'allOf': [{
        'type': 'object',
        'required': ['C', 'fit_intercept', 'max_iter', 'tol', 'early_stopping', 'validation_fraction', 'n_iter_no_change', 'shuffle', 'verbose', 'loss', 'n_jobs', 'random_state', 'warm_start', 'class_weight', 'average', 'n_iter'],
        'relevantToOptimizer': ['fit_intercept', 'max_iter', 'tol', 'shuffle', 'loss', 'n_iter'],
        'additionalProperties': False,
        'properties': {
            'C': {
                'type': 'number',
                'default': 1.0,
                'description': 'Maximum step size (regularization)'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether the intercept should be estimated or not'},
            'max_iter': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 10,
                    'maximumForOptimizer': 1000,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The maximum number of passes over the training data (aka epochs)'},
            'tol': {
                'anyOf': [{
                    'type': 'number',
                    'minimumForOptimizer': 1e-08,
                    'maximumForOptimizer': 0.01,
                    'distribution': 'loguniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The stopping criterion'},
            'early_stopping': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to use early stopping to terminate training when validation'},
            'validation_fraction': {
                'type': 'number',
                'default': 0.1,
                'description': 'The proportion of training data to set aside as validation set for early stopping'},
            'n_iter_no_change': {
                'type': 'integer',
                'default': 5,
                'description': 'Number of iterations with no improvement to wait before early stopping'},
            'shuffle': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether or not the training data should be shuffled after each epoch.'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'The verbosity level'},
            'loss': {
                'enum': ['epsilon_insensitive', 'huber', 'log', 'modified_huber', 'perceptron', 'squared_epsilon_insensitive', 'squared_hinge', 'squared_loss', 'hinge'],
                'default': 'hinge',
                'description': 'The loss function to be used: hinge: equivalent to PA-I in the reference paper'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of CPUs to use to do the OVA (One Versus All, for multi-class problems) computation'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The seed of the pseudo random number generator to use when shuffling the data'},
            'warm_start': {
                'type': 'boolean',
                'default': False,
                'description': 'When set to True, reuse the solution of the previous call to fit as initialization, otherwise, just erase the previous solution'},
            'class_weight': {
                'XXX TODO XXX': 'dict, {class_label: weight} or "balanced" or None, optional',
                'description': 'Preset for the class_weight fit parameter',
                'enum': ['balanced'],
                'default': 'balanced'},
            'average': {
                'anyOf': [{
                    'type': 'boolean'}, {
                    'type': 'integer'}],
                'default': False,
                'description': 'When set to True, computes the averaged SGD weights and stores the result in the ``coef_`` attribute'},
            'n_iter': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 5,
                    'maximumForOptimizer': 1000,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of passes over the training data (aka epochs)'},
        }}, {
        'XXX TODO XXX': 'Parameter: max_iter > only impacts the behavior in the fit method, and not the partial_fit'}, {
        'description': 'validation_fraction, only used if early_stopping is true',
        'anyOf': [{
            'type': 'object',
            'properties': {
                'validation_fraction': {
                    'enum': [0.1]},
            }}, {
            'type': 'object',
            'properties': {
                'early_stopping': {
                    'enum': [True]},
            }}]}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit linear model with Passive Aggressive algorithm.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values'},
        'coef_init': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The initial coefficients to warm-start the optimization.'},
        'intercept_init': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'The initial intercept to warm-start the optimization.'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted class label per sample.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict confidence scores for samples.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Confidence scores per (sample, class) combination',
    'laleType': 'Any',
    'XXX TODO XXX': 'array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.PassiveAggressiveClassifier#sklearn-linear_model-passiveaggressiveclassifier',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema},
}
lale.docstrings.set_docstrings(PassiveAggressiveClassifierImpl, _combined_schemas)
PassiveAggressiveClassifier = lale.operators.make_operator(PassiveAggressiveClassifierImpl, _combined_schemas)





from sklearn.manifold.locally_linear import LocallyLinearEmbedding as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class LocallyLinearEmbeddingImpl():

    def __init__(self, n_neighbors=5, n_components=2, reg=0.001, eigen_solver='auto', tol=1e-06, max_iter=100, method='standard', hessian_tol=0.0001, modified_tol=1e-12, neighbors_algorithm='auto', random_state=None, n_jobs=None):
        self._hyperparams = {
            'n_neighbors': n_neighbors,
            'n_components': n_components,
            'reg': reg,
            'eigen_solver': eigen_solver,
            'tol': tol,
            'max_iter': max_iter,
            'method': method,
            'hessian_tol': hessian_tol,
            'modified_tol': modified_tol,
            'neighbors_algorithm': neighbors_algorithm,
            'random_state': random_state,
            'n_jobs': n_jobs}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for LocallyLinearEmbedding    Locally Linear Embedding',
    'allOf': [{
        'type': 'object',
        'required': ['n_neighbors', 'n_components', 'reg', 'eigen_solver', 'tol', 'max_iter', 'method', 'hessian_tol', 'modified_tol', 'neighbors_algorithm', 'random_state', 'n_jobs'],
        'relevantToOptimizer': ['n_neighbors', 'n_components', 'eigen_solver', 'tol', 'max_iter', 'method', 'neighbors_algorithm'],
        'additionalProperties': False,
        'properties': {
            'n_neighbors': {
                'type': 'integer',
                'minimumForOptimizer': 5,
                'maximumForOptimizer': 20,
                'distribution': 'uniform',
                'default': 5,
                'description': 'number of neighbors to consider for each point.'},
            'n_components': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 256,
                'distribution': 'uniform',
                'default': 2,
                'description': 'number of coordinates for the manifold'},
            'reg': {
                'type': 'number',
                'default': 0.001,
                'description': 'regularization constant, multiplies the trace of the local covariance matrix of the distances.'},
            'eigen_solver': {
                'enum': ['auto', 'arpack', 'dense'],
                'default': 'auto',
                'description': 'auto : algorithm will attempt to choose the best method for input data  arpack : use arnoldi iteration in shift-invert mode'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 1e-06,
                'description': "Tolerance for 'arpack' method Not used if eigen_solver=='dense'."},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 100,
                'description': 'maximum number of iterations for the arpack solver'},
            'method': {
                'XXX TODO XXX': "string ('standard', 'hessian', 'modified' or 'ltsa')",
                'description': 'standard : use the standard locally linear embedding algorithm',
                'enum': ['ltsa', 'modified', 'standard'],
                'default': 'standard'},
            'hessian_tol': {
                'type': 'number',
                'default': 0.0001,
                'description': 'Tolerance for Hessian eigenmapping method'},
            'modified_tol': {
                'type': 'number',
                'default': 1e-12,
                'description': 'Tolerance for modified LLE method'},
            'neighbors_algorithm': {
                'enum': ['auto', 'brute', 'kd_tree', 'ball_tree'],
                'default': 'auto',
                'description': 'algorithm to use for nearest neighbors search, passed to neighbors.NearestNeighbors instance'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of parallel jobs to run'},
        }}, {
        'description': "hessian_tol, only used if method == 'hessian'",
        'anyOf': [{
            'type': 'object',
            'properties': {
                'hessian_tol': {
                    'enum': [0.0001]},
            }}, {
            'type': 'object',
            'properties': {
                'method': {
                    'enum': ['hessian']},
            }}]}, {
        'description': "modified_tol, only used if method == 'modified'",
        'anyOf': [{
            'type': 'object',
            'properties': {
                'modified_tol': {
                    'enum': [1e-12]},
            }}, {
            'type': 'object',
            'properties': {
                'method': {
                    'enum': ['modified']},
            }}]}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Compute the embedding vectors for data X',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'training set.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform new points into embedding space.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform new points into embedding space.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.manifold.LocallyLinearEmbedding#sklearn-manifold-locallylinearembedding',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(LocallyLinearEmbeddingImpl, _combined_schemas)
LocallyLinearEmbedding = lale.operators.make_operator(LocallyLinearEmbeddingImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lale operators with schemas.

Operators
=========

Estimators:

* lale.lib.lale. `BaselineClassifier`_
* lale.lib.lale. `BaselineRegressor`_
* lale.lib.lale. `GridSearchCV`_
* lale.lib.lale. `Hyperopt`_
* lale.lib.lale. `TopKVotingClassifier`_
* lale.lib.lale. `SMAC`_

Transformers:

* lale.lib.lale. `Batching`_
* lale.lib.lale. `ConcatFeatures`_
* lale.lib.lale. `NoOp`_
* lale.lib.lale. `Project`_
* lale.lib.lale. `SampleBasedVoting`_

Estimators and transformers:

* lale.lib.lale. `Both`_
* lale.lib.lale. `IdentityWrapper`_
* lale.lib.lale. `Observing`_

.. _`BaselineClassifier`: lale.lib.lale.baseline_classifier.html
.. _`BaselineRegressor`: lale.lib.lale.baseline_regressor.html
.. _`GridSearchCV`: lale.lib.lale.grid_search_cv.html
.. _`Hyperopt`: lale.lib.lale.hyperopt.html
.. _`TopKVotingClassifier`: lale.lib.lale.topk_voting_classifier.html
.. _`SMAC`: lale.lib.lale.smac.html
.. _`Batching`: lale.lib.lale.batching.html
.. _`ConcatFeatures`: lale.lib.lale.concat_features.html
.. _`NoOp`: lale.lib.lale.no_op.html
.. _`Project`: lale.lib.lale.project.html
.. _`SampleBasedVoting`: lale.lib.lale.sample_based_voting.html
.. _`Both`: lale.lib.lale.both.html
.. _`IdentityWrapper`: lale.lib.lale.identity_wrapper.html
.. _`Observing`: lale.lib.lale.observing.html
"""

#estimators
from .baseline_classifier import BaselineClassifier
from .baseline_regressor import BaselineRegressor
from .grid_search_cv import GridSearchCV
from .hyperopt import Hyperopt
from .topk_voting_classifier import TopKVotingClassifier

import warnings
with warnings.catch_warnings(record=True) as w:
    try:
        from .smac import SMAC
    except ImportError:
        pass

#transformers
from .batching import Batching
from .concat_features import ConcatFeatures
from .no_op import NoOp
from .project import Project
from .sample_based_voting import SampleBasedVoting

#estimators and transformers
from .both import Both
from .identity_wrapper import IdentityWrapper
from .observing import Observing



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import autoai_libs.cognito.transforms.transform_utils
import lale.docstrings
import lale.operators

class FS2Impl():
    def __init__(self, cols_ids_must_keep, additional_col_count_to_keep, ptype, eval_algo):
        self._hyperparams = {
            'cols_ids_must_keep': cols_ids_must_keep,
            'additional_col_count_to_keep': additional_col_count_to_keep,
            'ptype': ptype,
            'eval_algo': eval_algo}
        self._wrapped_model = autoai_libs.cognito.transforms.transform_utils.FS2(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        result = self._wrapped_model.transform(X)
        return result

_hyperparams_schema = {
    'allOf': [{
        'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'additionalProperties': False,
        'required': ['cols_ids_must_keep', 'additional_col_count_to_keep', 'ptype', 'eval_algo'],
        'relevantToOptimizer': [],
        'properties': {
            'cols_ids_must_keep': {
                'description': 'Serial numbers of the columns that must be kept irrespective of their feature importance.',
                'laleType':'Any',#Found a value `range(0, 20) in fs1` 
                'default': []},
            'additional_col_count_to_keep': {
                'description': 'How many columns need to be retained.',
                'type': 'integer',
                'minimum': 0},
            'ptype': {
                'description': 'Problem type.',
                'enum': ['classification', 'regression'],
                'default': 'classification'},
            'eval_algo': {
                'description': 'A supervised model where fit() sets `feature_importances_`.',
                'laleType': 'Any',
                'default': None}}}]}

_input_fit_schema = {
    'type': 'object',
    'required': ['X', 'y'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'laleType': 'Any'}}},
        'y': {
            'type': 'array',
            'items': {'laleType': 'Any'},
            'description': 'Target values.'}}}

_input_transform_schema = {
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'laleType': 'Any'}}}}}

_output_transform_schema = {
    'description': 'Features; the outer array is over samples.',
    'type': 'array',
    'items': {'type': 'array', 'items': {'laleType': 'Any'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """Operator from `autoai_libs`_. Feature selection, type 2.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.fs2.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(FS2Impl, _combined_schemas)

FS2 = lale.operators.make_operator(FS2Impl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import numpy as np
import pandas as pd
import os
import re
import sys
import time
import traceback
import scipy.sparse
import importlib
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score, log_loss
from sklearn.metrics.scorer import check_scoring
from sklearn.utils.metaestimators import _safe_split
import copy
import logging
import h5py
from typing import Any, Dict, List, Optional, Union
import lale.datasets.data_schemas

try:
    import torch
    torch_installed=True
except ImportError:
    torch_installed=False

logger = logging.getLogger(__name__)

class NestedHyperoptSpace():
    sub_space:Any
    def __init__(self, sub_space):
        self.sub_space = sub_space
    def __str__(self):
        return str(self.sub_space)
    #define __repr__ so that __str__ gets invoked while printing lists and dictionaries
    __repr__ = __str__

def assignee_name(level=1) -> Optional[str]:
    tb = traceback.extract_stack()
    file_name, line_number, function_name, text = tb[-(level+2)]
    tree = ast.parse(text, file_name)
    assert isinstance(tree, ast.Module)
    if len(tree.body) == 1 and isinstance(tree.body[0], ast.Assign):
        lhs = tree.body[0].targets
        if len(lhs) == 1 and isinstance(lhs[0], ast.Name):
            return lhs[0].id
    return None

def data_to_json(data, subsample_array:bool=True) -> Union[list, dict]:
    if type(data) is tuple:
        # convert to list
        return [data_to_json(elem, subsample_array) for elem in data]
    if type(data) is list:
        return [data_to_json(elem, subsample_array) for elem in data]
    elif type(data) is dict:
        return {key: data_to_json(data[key], subsample_array) for key in data}
    elif isinstance(data, np.ndarray):
        return ndarray_to_json(data, subsample_array)
    elif type(data) is scipy.sparse.csr_matrix:
        return ndarray_to_json(data.toarray(), subsample_array)
    elif isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        np_array = data.values
        return ndarray_to_json(np_array, subsample_array)
    elif torch_installed and isinstance(data, torch.Tensor):
        np_array = data.detach().numpy()
        return ndarray_to_json(np_array, subsample_array)
    else:
        return data

def is_empty_dict(val) -> bool:
    return isinstance(val, dict) and len(val) == 0

def dict_without(orig_dict: Dict[str, Any], key: str) -> Dict[str, Any]:
    return {k: orig_dict[k] for k in orig_dict if k != key}

def ndarray_to_json(arr, subsample_array:bool=True) -> Union[list, dict]:
    #sample 10 rows and no limit on columns
    if subsample_array:
        num_subsamples = [10, np.iinfo(np.int).max, np.iinfo(np.int).max]
    else:
        num_subsamples = [np.iinfo(np.int).max, np.iinfo(np.int).max, np.iinfo(np.int).max]
    def subarray_to_json(indices):
        if len(indices) == len(arr.shape):
            if isinstance(arr[indices], bool) or \
               isinstance(arr[indices], int) or \
               isinstance(arr[indices], float) or \
               isinstance(arr[indices], str):
                return arr[indices]
            elif np.issubdtype(arr.dtype, np.bool_):
                return bool(arr[indices])
            elif np.issubdtype(arr.dtype, np.integer):
                return int(arr[indices])
            elif np.issubdtype(arr.dtype, np.number):
                return float(arr[indices])
            elif arr.dtype.kind in ['U', 'S', 'O']:
                return str(arr[indices])
            else:
                raise ValueError(f'Unexpected dtype {arr.dtype}, '
                                 f'kind {arr.dtype.kind}, '
                                 f'type {type(arr[indices])}.')
        else:
            assert len(indices) < len(arr.shape)
            return [subarray_to_json(indices + (i,))
                    for i in range(min(num_subsamples[len(indices)], arr.shape[len(indices)]))]
    return subarray_to_json(())

def split_with_schemas(estimator, all_X, all_y, indices, train_indices=None):
    subset_X, subset_y = _safe_split(
        estimator, all_X, all_y, indices, train_indices)
    if hasattr(all_X, 'json_schema'):
        n_rows = subset_X.shape[0]
        schema = {
            'type': 'array', 'minItems': n_rows, 'maxItems': n_rows,
            'items': all_X.json_schema['items']}
        lale.datasets.data_schemas.add_schema(subset_X, schema)
    if hasattr(all_y, 'json_schema'):
        n_rows = subset_y.shape[0]
        schema = {
            'type': 'array', 'minItems': n_rows, 'maxItems': n_rows,
            'items': all_y.json_schema['items']}
        lale.datasets.data_schemas.add_schema(subset_y, schema)
    return subset_X, subset_y

def cross_val_score_track_trials(estimator, X, y=None, scoring=accuracy_score, cv=5, args_to_scorer=None):
    """
    Use the given estimator to perform fit and predict for splits defined by 'cv' and compute the given score on 
    each of the splits.

    Parameters
    ----------

    estimator: A valid sklearn_wrapper estimator
    X, y: Valid data and target values that work with the estimator
    scoring: string or a scorer object created using 
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer.
        A string from sklearn.metrics.SCORERS.keys() can be used or a scorer created from one of 
        sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics).
        A completely custom scorer object can be created from a python function following the example at 
        https://scikit-learn.org/stable/modules/model_evaluation.html
        The metric has to return a scalar value,
    cv: an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
        Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
        Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.
    args_to_scorer: A dictionary of additional keyword arguments to pass to the scorer. 
                Used for cases where the scorer has a signature such as ``scorer(estimator, X, y, **kwargs)``.
    Returns
    -------
        cv_results: a list of scores corresponding to each cross validation fold
    """
    if isinstance(cv, int):
        cv = StratifiedKFold(cv)

    if args_to_scorer is None:
        args_to_scorer={}
    scorer = check_scoring(estimator, scoring=scoring)
    cv_results:List[float] = []
    log_loss_results = []
    time_results = []
    for train, test in cv.split(X, y):
        X_train, y_train = split_with_schemas(estimator, X, y, train)
        X_test, y_test = split_with_schemas(estimator, X, y, test, train)
        start = time.time()
        #Not calling sklearn.base.clone() here, because:
        #  (1) For Lale pipelines, clone() calls the pipeline constructor
        #      with edges=None, so the resulting topology is incorrect.
        #  (2) For Lale individual operators, the fit() method already
        #      clones the impl object, so cloning again is redundant.
        trained = estimator.fit(X_train, y_train)
        score_value  = scorer(trained, X_test, y_test, **args_to_scorer)
        execution_time = time.time() - start
        # not all estimators have predict probability
        try:
            y_pred_proba = trained.predict_proba(X_test)
            logloss = log_loss(y_true=y_test, y_pred=y_pred_proba)
            log_loss_results.append(logloss)
        except BaseException:
            logger.debug("Warning, log loss cannot be computed")
        cv_results.append(score_value)
        time_results.append(execution_time)
    result = np.array(cv_results).mean(), np.array(log_loss_results).mean(), np.array(execution_time).mean()
    return result


def cross_val_score(estimator, X, y=None, scoring=accuracy_score, cv=5):
    """
    Use the given estimator to perform fit and predict for splits defined by 'cv' and compute the given score on
    each of the splits.
    :param estimator: A valid sklearn_wrapper estimator
    :param X, y: Valid data and target values that work with the estimator
    :param scoring: a scorer object from sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics)
             Default value is accuracy_score.
    :param cv: an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
        Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
        Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.
    :return: cv_results: a list of scores corresponding to each cross validation fold
    """
    if isinstance(cv, int):
        cv = StratifiedKFold(cv)

    cv_results = []
    for train, test in cv.split(X, y):
        X_train, y_train = split_with_schemas(estimator, X, y, train)
        X_test, y_test = split_with_schemas(estimator, X, y, test, train)
        trained_estimator = estimator.fit(X_train, y_train)
        predicted_values = trained_estimator.predict(X_test)
        cv_results.append(scoring(y_test, predicted_values))

    return cv_results

def create_operator_using_reflection(class_name, operator_name, param_dict):
    instance = None
    if class_name is not None:
        class_name_parts = class_name.split(".")
        assert(len(class_name_parts)) >1, "The class name needs to be fully qualified, i.e. module name + class name"
        module_name = ".".join(class_name_parts[0:-1])
        class_name = class_name_parts[-1]

        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        
        if param_dict is None:
            instance = class_.create(name=operator_name)
        else:
            instance = class_.create(name=operator_name, kwargs = param_dict)
    return instance

def create_individual_op_using_reflection(class_name, operator_name, param_dict):
    instance = None
    if class_name is not None:
        class_name_parts = class_name.split(".")
        assert(len(class_name_parts)) >1, "The class name needs to be fully qualified, i.e. module name + class name"
        module_name = ".".join(class_name_parts[0:-1])
        class_name = class_name_parts[-1]

        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        
        if param_dict is None:
            instance = class_()
        else:
            instance = class_(**param_dict)
    return instance

def to_graphviz(lale_operator: 'lale.operators.Operator', ipython_display:bool=True, call_depth:int=1, **dot_graph_attr):
    import lale.visualize
    if not isinstance(lale_operator, lale.operators.Operator):
        raise TypeError("The input to to_graphviz needs to be a valid LALE operator.")
    jsn = lale.json_operator.to_json(lale_operator, call_depth=call_depth+1)
    dot = lale.visualize.json_to_graphviz(jsn, ipython_display, dot_graph_attr)
    return dot

def println_pos(message, out_file=sys.stdout):
    tb = traceback.extract_stack()[-2]
    match = re.search(r'<ipython-input-([0-9]+)-', tb[0])
    if match:
        pos = 'notebook cell [{}] line {}'.format(match[1], tb[1])
    else:
        pos = '{}:{}'.format(tb[0], tb[1])
    strtime = time.strftime('%Y-%m-%d_%H-%M-%S')
    to_log = '{}: {} {}'.format(pos, strtime, message)
    print(to_log, file=out_file)
    if match:
        os.system('echo {}'.format(to_log))

def instantiate_from_hyperopt_search_space(obj_hyperparams, new_hyperparams):
    if isinstance(new_hyperparams, NestedHyperoptSpace):
        sub_params = new_hyperparams.sub_space

        sub_op = obj_hyperparams
        if isinstance(sub_op, list):
            if len(sub_op)==1:
                sub_op = sub_op[0]
            else:
                step_index, step_params = list(sub_params)[0]
                if step_index < len(sub_op):
                    sub_op = sub_op[step_index]
                    sub_params = step_params

        return create_instance_from_hyperopt_search_space(sub_op, sub_params)

    elif isinstance(new_hyperparams, (list, tuple)):
        assert isinstance(obj_hyperparams, (list, tuple))
        params_len = len(new_hyperparams)
        assert params_len == len(obj_hyperparams)
        res:Optional[List[Any]] = None

        for i in range(params_len):
            nhi = new_hyperparams[i]
            ohi = obj_hyperparams[i]
            updated_params = instantiate_from_hyperopt_search_space(ohi, nhi)
            if updated_params is not None:
                if res is None:
                    res = list(new_hyperparams)
                res[i] = updated_params
        if res is not None:
            if isinstance(obj_hyperparams, tuple):
                return tuple(res)
            else:
                return res
        # workaround for what seems to be a hyperopt bug
        # where hyperopt returns a tuple even though the
        # hyperopt search space specifies a list
        is_obj_tuple = isinstance(obj_hyperparams, tuple)
        is_new_tuple = isinstance(new_hyperparams, tuple)
        if is_obj_tuple != is_new_tuple:
            if is_obj_tuple:
                return tuple(new_hyperparams)
            else:
                return list(new_hyperparams)
        return None

    elif isinstance(new_hyperparams, dict):
        assert isinstance(obj_hyperparams, dict)

        for k,sub_params in new_hyperparams.items():
            if k in obj_hyperparams:
                sub_op = obj_hyperparams[k]
                updated_params = instantiate_from_hyperopt_search_space(sub_op, sub_params)
                if updated_params is not None:
                    new_hyperparams[k] = updated_params
        return None
    else:
        return None

def create_instance_from_hyperopt_search_space(lale_object, hyperparams):
    '''
    Hyperparams is a n-tuple of dictionaries of hyper-parameters, each
    dictionary corresponds to an operator in the pipeline
    '''
    #lale_object can either be an individual operator, a pipeline or an operatorchoice
    #Validate that the number of elements in the n-tuple is the same
    #as the number of steps in the current pipeline

    from lale.operators import IndividualOp
    from lale.operators import BasePipeline
    from lale.operators import TrainablePipeline
    from lale.operators import Operator
    from lale.operators import OperatorChoice
    if isinstance(lale_object, IndividualOp):
        new_hyperparams:Dict[str,Any] = dict_without(hyperparams, 'name')
        if lale_object._hyperparams is not None:
            obj_hyperparams = dict(lale_object._hyperparams)
        else:
            obj_hyperparams = {}

        for k,sub_params in new_hyperparams.items():
            if k in obj_hyperparams:
                sub_op = obj_hyperparams[k]
                updated_params = instantiate_from_hyperopt_search_space(sub_op, sub_params)
                if updated_params is not None:
                    new_hyperparams[k] = updated_params

        all_hyperparams = {**obj_hyperparams, **new_hyperparams}
        return lale_object(**all_hyperparams)
    elif isinstance(lale_object, BasePipeline):
        steps = lale_object.steps()
        if len(hyperparams) != len(steps):
            raise ValueError('The number of steps in the hyper-parameter space does not match the number of steps in the pipeline.')
        op_instances = []
        edges = lale_object.edges()
        #op_map:Dict[PlannedOpType, TrainableOperator] = {}
        op_map = {}
        for op_index, sub_params in enumerate(hyperparams):
            sub_op = steps[op_index]
            op_instance = create_instance_from_hyperopt_search_space(sub_op, sub_params)
            assert isinstance(sub_op, OperatorChoice) or sub_op.class_name() == op_instance.class_name(), f'sub_op {sub_op.class_name()}, op_instance {op_instance.class_name()}'
            op_instances.append(op_instance)
            op_map[sub_op] = op_instance

        #trainable_edges:List[Tuple[TrainableOperator, TrainableOperator]]
        try:
            trainable_edges = [(op_map[x], op_map[y]) for (x, y) in edges]
        except KeyError as e:
            raise ValueError("An edge was found with an endpoint that is not a step (" + str(e) + ")")

        return TrainablePipeline(op_instances, trainable_edges, ordered=True)
    elif isinstance(lale_object, OperatorChoice):
        #Hyperopt search space for an OperatorChoice is generated as a dictionary with a single element
        #corresponding to the choice made, the only key is the index of the step and the value is 
        #the params corresponding to that step.
        step_index:int
        choices = lale_object.steps()

        if len(choices)==1:
            step_index = 0
        else:
            step_index, hyperparams = list(hyperparams.items())[0]
        step_object = choices[step_index]
        return create_instance_from_hyperopt_search_space(step_object, hyperparams)
        
def import_from_sklearn_pipeline(sklearn_pipeline, fitted=True):
    #For all pipeline steps, identify equivalent lale wrappers if present,
    #if not, call make operator on sklearn classes and create a lale pipeline.

    def get_equivalent_lale_op(sklearn_obj, fitted):
        module_names = ["lale.lib.sklearn", "lale.lib.autoai_libs"]
        from lale.operators import make_operator, TrainedIndividualOp

        lale_wrapper_found = False
        class_name = sklearn_obj.__class__.__name__
        for module_name in module_names:
            module = importlib.import_module(module_name)
            try:
                class_ = getattr(module, class_name)
                lale_wrapper_found  = True
                break
            except AttributeError:
                continue
        else:
            class_ = make_operator(sklearn_obj, name=class_name)

        if not fitted:#If fitted is False, we do not want to return a Trained operator.
            lale_op = class_
        else:
            lale_op = TrainedIndividualOp(class_._name, class_._impl, class_._schemas)
        class_ = lale_op(**sklearn_obj.get_params())
        if lale_wrapper_found:
            class_._impl_instance()._wrapped_model =  copy.deepcopy(sklearn_obj)
        else:# If there is no lale wrapper, there is no _wrapped_model
            class_._impl = copy.deepcopy(sklearn_obj) 
        return class_

    from sklearn.pipeline import FeatureUnion, Pipeline
    from sklearn.base import BaseEstimator
    from lale.operators import make_pipeline, make_union
    if isinstance(sklearn_pipeline, Pipeline):
        nested_pipeline_steps = sklearn_pipeline.named_steps.values()
        nested_pipeline_lale_objects = [import_from_sklearn_pipeline(nested_pipeline_step, fitted=fitted) for nested_pipeline_step in nested_pipeline_steps]
        lale_op_obj = make_pipeline(*nested_pipeline_lale_objects)
    elif isinstance(sklearn_pipeline, FeatureUnion):
        transformer_list = sklearn_pipeline.transformer_list
        concat_predecessors = [import_from_sklearn_pipeline(transformer[1], fitted=fitted) for transformer in transformer_list]
        lale_op_obj = make_union(*concat_predecessors)
    else:
        lale_op_obj = get_equivalent_lale_op(sklearn_pipeline, fitted=fitted)
    return lale_op_obj

class val_wrapper():
    """This is used to wrap values that cause problems for hyper-optimizer backends
    lale will unwrap these when given them as the value of a hyper-parameter"""

    def __init__(self, base):
        self._base = base

    def unwrap_self(self):
        return self._base

    @classmethod
    def unwrap(cls, obj):
        if isinstance(obj, cls):
            return cls.unwrap(obj.unwrap_self())
        else:
            return obj

def append_batch(data, batch_data):
    if data is None:
        return batch_data
    elif isinstance(data, np.ndarray):
        if isinstance(batch_data, np.ndarray):
            if len(data.shape) == 1 and len(batch_data.shape) == 1:
                return np.concatenate([data, batch_data])
            else:
                return np.vstack((data, batch_data))
    elif isinstance(data, tuple):
        X, y = data
        if isinstance(batch_data, tuple):
            batch_X, batch_y = batch_data
            X = append_batch(X, batch_X)
            y = append_batch(y, batch_y)
            return X, y
    elif torch_installed and isinstance(data, torch.Tensor):
        if isinstance(batch_data, torch.Tensor):
            return torch.cat((data, batch_data))
    elif isinstance(data, h5py.File):
        if isinstance(batch_data, tuple):
            batch_X, batch_y = batch_data
            
    #TODO:Handle dataframes

def create_data_loader(X, y = None, batch_size = 1):
    from lale.util.numpy_to_torch_dataset import NumpyTorchDataset
    from lale.util.hdf5_to_torch_dataset import HDF5TorchDataset
    from torch.utils.data import DataLoader, TensorDataset
    from lale.util.batch_data_dictionary_dataset import BatchDataDict
    import torch
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()
        if isinstance(y, pd.Series):
            y = y.to_numpy()
    elif isinstance(X, np.ndarray):
        #unfortunately, NumpyTorchDataset won't accept a subclass of np.ndarray
        if isinstance(X, lale.datasets.data_schemas.NDArrayWithSchema):
            X = X.view(np.ndarray)
        if isinstance(y, lale.datasets.data_schemas.NDArrayWithSchema):
            y = y.view(np.ndarray)
        dataset = NumpyTorchDataset(X, y)
    elif isinstance(X, str):#Assume that this is path to hdf5 file
        dataset = HDF5TorchDataset(X)
    elif isinstance(X, BatchDataDict):
        dataset = X
        def my_collate_fn(batch):
            return batch[0]#because BatchDataDict's get_item returns a batch, so no collate is required.
        return DataLoader(dataset, batch_size=1, collate_fn=my_collate_fn)
    elif isinstance(X, dict): #Assumed that it is data indexed by batch number
        return [X]
    elif isinstance(X, torch.Tensor) and y is not None:
        if isinstance(y, np.ndarray):
            y = torch.from_numpy(y)
        dataset = TensorDataset(X, y)
    elif isinstance(X, torch.Tensor):
        dataset = TensorDataset(X)
    else:
        raise TypeError("Can not create a data loader for a dataset with type {}".format(type(X)))
    return DataLoader(dataset, batch_size=batch_size)

def write_batch_output_to_file(file_obj, file_path, total_len, batch_idx, batch_X, batch_y, batch_out_X, batch_out_y):
    if file_obj is None and file_path is None:
        raise ValueError("Only one of the file object or file path can be None.")
    if file_obj is None:
        file_obj = h5py.File(file_path, 'w')
        #estimate the size of the dataset based on the first batch output size
        transform_ratio = int(len(batch_out_X)/len(batch_X))
        if len(batch_out_X.shape) == 1:
            h5_data_shape = (transform_ratio*total_len, )
        if len(batch_out_X.shape) == 2:
            h5_data_shape = (transform_ratio*total_len, batch_out_X.shape[1])
        elif len(batch_out_X.shape) == 3:
            h5_data_shape = (transform_ratio*total_len, batch_out_X.shape[1], batch_out_X.shape[2])
        dataset = file_obj.create_dataset(name='X', shape=h5_data_shape, chunks=True, compression="gzip")
        if batch_out_y is None and batch_y is not None:
            batch_out_y = batch_y
        if batch_out_y is not None:
            if len(batch_out_y.shape) == 1:
                h5_labels_shape = (transform_ratio*total_len, )
            elif len(batch_out_y.shape) == 2:
                h5_labels_shape = (transform_ratio*total_len, batch_out_y.shape[1])
            dataset = file_obj.create_dataset(name='y', shape=h5_labels_shape, chunks=True, compression="gzip")
    dataset = file_obj['X']
    dataset[batch_idx*len(batch_out_X):(batch_idx+1)*len(batch_out_X)] = batch_out_X
    if batch_out_y is not None or batch_y is not None:
        labels = file_obj['y']
        if batch_out_y is not None:
            labels[batch_idx*len(batch_out_y):(batch_idx+1)*len(batch_out_y)] = batch_out_y
        else:
            labels[batch_idx*len(batch_y):(batch_idx+1)*len(batch_y)] = batch_y
    return file_obj

def best_estimator(obj):
    """
    .. deprecated:: 0.3.3
       The best_estimator(obj) function has been replaced by the
       obj.get_pipeline() method.
    """
    return obj.get_pipeline()



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
import numpy as np

class BaselineRegressorImpl:
    def __init__(self):
        pass

    def fit(self, X, y):
        self._average_label = np.average(y)
        return self

    def predict(self, X):
        result = np.full((X.shape[0],), self._average_label)
        return result

_hyperparams_schema = {
    'allOf': [
    {   'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'relevantToOptimizer': [],
        'additionalProperties': False}]}

_input_fit_schema = {
    'required': ['X', 'y'],
    'type': 'object',
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array'}},
        'y': {
            'description': 'Target values.',
            'type': 'array', 'items': {'type': 'number'}}}}

_input_predict_schema = {
    'type': 'object',
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array'}}}}

_output_predict_schema = {
    'description': 'Predicted values per sample.',
    'type': 'array', 'items': {'type': 'number'}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Baseline regressor always predicts the average target value.',
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.baseline_regressor.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema}}

lale.docstrings.set_docstrings(BaselineRegressorImpl, _combined_schemas)

BaselineRegressor = lale.operators.make_operator(BaselineRegressorImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import numpy

from typing import Any, Dict, List, Set, Iterable, Iterator, Optional, Tuple, Union
from hyperopt import hp
from hyperopt.pyll import scope
from lale.util.VisitorMeta import AbstractVisitorMeta
from lale.search.PGO import FrequencyDistribution
import os

PGO_input_type = Union[FrequencyDistribution, Iterable[Tuple[Any, int]], None]
class SearchSpace(metaclass=AbstractVisitorMeta):
    def __init__(self, default:Optional[Any]=None):
        self._default = default

    _default:Optional[Any]

    def default(self)->Optional[Any]:
        """Return an optional default value, if None.
            if not None, the default value should be in the
            search space
        """
        return self._default

class SearchSpaceEmpty(SearchSpace):
    def __init__(self):
        super(SearchSpaceEmpty, self).__init__()

    def __str__(sefl):
        return "***EMPTY***"

class SearchSpacePrimitive(SearchSpace):
    def __init__(self, default:Optional[Any]=None):
        super(SearchSpacePrimitive, self).__init__(default=default)

class SearchSpaceEnum(SearchSpacePrimitive):
    pgo:Optional[FrequencyDistribution]
    vals:List[Any]
    def __init__(self, vals:Iterable[Any], 
                 pgo:PGO_input_type=None, 
                 default:Optional[Any]=None):
        super(SearchSpaceEnum, self).__init__(default=default)
        self.vals = sorted(vals, key=str)

        if pgo is None or isinstance(pgo, FrequencyDistribution):
            self.pgo = pgo
        else:
            self.pgo = FrequencyDistribution.asEnumValues(pgo, self.vals)

    def __str__(self):
        return "<" + ",".join(map(str, self.vals)) + ">"

class SearchSpaceConstant(SearchSpaceEnum):
    def __init__(self, v, pgo:PGO_input_type=None):
        super(SearchSpaceConstant, self).__init__([v], pgo=pgo, default=v)

    def __str__(self):
        return str(self.vals[0])

class SearchSpaceBool(SearchSpaceEnum):
    def __init__(self, pgo:PGO_input_type=None, default:Optional[Any]=None):
        super(SearchSpaceBool, self).__init__([True, False], pgo=pgo, default=default)

class SearchSpaceNumber(SearchSpacePrimitive):
    minimum:Optional[float]
    exclusiveMinumum:bool
    maximum:Optional[float]
    exclusiveMaximum:bool
    discrete:bool
    distribution:str
    pgo:Optional[FrequencyDistribution]

    def __init__(self, 
                 minimum=None, 
                 exclusiveMinimum:bool=False, 
                 maximum=None, 
                 exclusiveMaximum:bool=False, 
                 discrete:bool=False, 
                 distribution="uniform",
                 pgo:PGO_input_type=None,
                 default:Optional[Any]=None) -> None:
        super(SearchSpaceNumber, self).__init__(default=default)
        self.minimum = minimum
        self.exclusiveMinimum = exclusiveMinimum
        self.maximum = maximum
        self.exclusiveMaximum = exclusiveMaximum
        self.distribution = distribution
        self.discrete = discrete
        if pgo is None or isinstance(pgo, FrequencyDistribution):
            self.pgo = pgo
        else:
            if discrete:
                self.pgo = FrequencyDistribution.asIntegerValues(pgo, inclusive_min=self.getInclusiveMin(), inclusive_max=self.getInclusiveMax())
            else:
                self.pgo = FrequencyDistribution.asFloatValues(pgo, inclusive_min=self.getInclusiveMin(), inclusive_max=self.getInclusiveMax())


    def getInclusiveMax(self):
        """ Return the maximum as an inclusive maximum (exclusive maxima are adjusted accordingly)
        """
        max = self.maximum
        if self.exclusiveMaximum:
            if self.discrete:
                max = max - 1
            else:
                max = numpy.nextafter(max, float('-inf'))
        return max

    def getInclusiveMin(self):
        """ Return the maximum as an inclusive minimum (exclusive minima are adjusted accordingly)
        """
        min = self.minimum
        if self.exclusiveMinimum:
            if self.discrete:
                min = min + 1
            else:
                min = numpy.nextafter(min, float('+inf'))
        return min

    def __str__(self):
        ret:str = ""
        if self.exclusiveMinimum or self.minimum is None:
             ret += "("
        else:

            ret += "["
        if self.discrete:
            ret += '\u2308'

        if self.minimum is None:
            ret += '\u221E'
        else:
            ret += str(self.minimum)

        if not self.distribution or self.distribution == "uniform" or self.distribution == "integer":
            ret += ","
        elif self.distribution == "loguniform":
            ret += ",<log>,"
        else:
            ret += ",<" + self.distribution + ">,"

        if self.maximum is None:
            ret += '\u221E'
        else:
            ret += str(self.maximum)

        if self.discrete:
            ret += '\u2309'
        if self.exclusiveMaximum or self.maximum is None:
             ret += ")"
        else:
            ret += "]"
        return ret

class SearchSpaceArray(SearchSpace):
    def __init__(self, prefix:Optional[List[SearchSpace]], minimum:int=0, *, maximum:int, additional:Optional[SearchSpace]=None, is_tuple=False) -> None:
        super(SearchSpaceArray, self).__init__()
        self.minimum = minimum
        self.maximum = maximum
        self.prefix = prefix
        self.additional = additional
        self.is_tuple = is_tuple
    
    def __str__(self):
        ret:str = ""
        ret += f"Array<{self.minimum}, {self.maximum}>"
        if self.is_tuple:
            ret += "("
        else:
            ret += "["

        if self.prefix is not None:
            ret += ",".join(map(str,self.prefix))
            if self.additional is not None:
                ret += ","
        if self.additional is not None:
            ret += "...,"
            ret += str(self.additional)

        if self.is_tuple:
            ret += ")"
        else:
            ret += "]"
        return ret

    def items(self, max:Optional[int]=None)->Iterable[SearchSpace]:
        prefix_len:int
        if self.prefix is not None:
            prefix_len = len(self.prefix)
        else:
            prefix_len = 0

        num_elts = self.maximum
        if max is not None:
            num_elts = min(num_elts, max)

        for i in range(num_elts):
            if self.prefix is not None and i < prefix_len:
                yield self.prefix[i]
            else:
                if self.additional is not None:
                    yield self.additional

class SearchSpaceObject(SearchSpace):
    def __init__(self, longName:str, keys:List[str], choices:Iterable[Any]) -> None:
        super(SearchSpaceObject, self).__init__()
        self.longName = longName
        self.keys = keys
        self.choices = choices

    def __str__(self):
        ret:str = ""
        ret += f"Object<{self.longName}>["

        choice_strs:List[str] = []
        for c in self.choices:
            opts:List[str] = []
            for k,v in zip(self.keys, c):
                opts.append(k + "->" + str(v))
            l = ";".join(opts)
            choice_strs.append("{" + l + "}")

        ret += ",".join(choice_strs) + "]"

        return ret

class SearchSpaceSum(SearchSpace):
    sub_spaces:List[SearchSpace]
    def __init__(self, 
                 sub_spaces:List[SearchSpace], 
                 default:Optional[Any]=None):
        super(SearchSpaceSum, self).__init__(default=default)
        self.sub_spaces = sub_spaces

    def __str__(self):
        ret:str = "\u2211["
        ret += "|".join(map(str, self.sub_spaces))
        ret += "]"
        return ret

class SearchSpaceOperator(SearchSpace):
    sub_space:SearchSpace
    def __init__(self,
                 sub_space:SearchSpace,
                 default:Optional[Any]=None):
        super(SearchSpaceOperator, self).__init__(default=default)
        self.sub_space = sub_space

    def __str__(self):
        ret:str = "\u00AB"
        ret += str(self.sub_space)
        ret += "\u00BB"
        return ret

class SearchSpaceProduct(SearchSpace):
    sub_spaces:List[Tuple[str, SearchSpace]]
    def __init__(self, 
                 sub_spaces:List[Tuple[str,SearchSpace]],
                 default:Optional[Any]=None):
        super(SearchSpaceProduct, self).__init__(default=default)
        self.sub_spaces = sub_spaces
    
    def get_indexed_spaces(self)->Iterable[Tuple[str, int, SearchSpace]]:
        indices:Dict[str,int]={}
        def make_indexed(name:str)->Tuple[str,int]:
            idx = 0
            if name in indices:
                idx = indices[name] + 1
                indices[name] = idx
            else:
                indices[name] = 0
            return (name, idx)
        def enhance_tuple(x:Tuple[str,int], space:SearchSpace)->Tuple[str,int,SearchSpace]:
            return (x[0], x[1], space)
        
        return [enhance_tuple(make_indexed(name), space) for name,space in self.sub_spaces]

    def __str__(self):
        ret:str = "\u220F{"

        parts:List[str] = []
        for k,v in self.sub_spaces:
            parts.append(k + "->" + str(v))
        ret = ";".join(parts)

        ret += "}"
        return ret

# for debugging
_print_search_space_env_options:Optional[Set[str]] = None

def _get_print_search_space_options()->Set[str]:
    global _print_search_space_env_options
    options:Set[str]
    if _print_search_space_env_options is None:
        debug = os.environ.get("LALE_PRINT_SEARCH_SPACE", None)
        if debug is None:
            options = set()
        else:
            options_raw = debug.split(",")
            options = set(s.strip().lower() for s in options_raw)
        _print_search_space_env_options = options
    else:
        options = _print_search_space_env_options
    return options

def should_print_search_space(*s:str):
    options:Set[str] = _get_print_search_space_options()
    for x in s:
        if x in options:
            return True
    return False
  



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
import sklearn.linear_model

_input_fit_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X', 'y'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}},
    'y': {
      'description': 'Target class labels; the array is over samples.',
        'anyOf': [
            {'type': 'array', 'items': {'type': 'number'}},
            {'type': 'array', 'items': {'type': 'string'}}]}}}

_input_predict_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted class label per sample.',
    'anyOf': [
        {'type': 'array', 'items': {'type': 'number'}},
        {'type': 'array', 'items': {'type': 'string'}}]}

_input_predict_proba_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Probability of the sample for each class in the model.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'}}}

_input_decision_function_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_decision_function_schema = {
    'description': 'Confidence scores for samples for each class in the model.',
    'anyOf': [
    {   'description': 'In the multi-way case, score per (sample, class) combination.',
        'type': 'array',
        'items': {'type': 'array', 'items': {'type': 'number'}}},
    {   'description': 'In the binary case, score for `self._classes[1]`.',
        'type': 'array',
        'items': {'type': 'number'}}]}

_hyperparams_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Hyperparameter schema.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'required': [
        'penalty', 'dual', 'tol', 'C', 'fit_intercept', 'intercept_scaling',
        'class_weight', 'random_state', 'solver', 'max_iter', 'multi_class',
        'verbose', 'warm_start', 'n_jobs'],
      'relevantToOptimizer': [
        'penalty', 'dual', 'tol', 'C', 'fit_intercept', 'class_weight',
        'solver', 'multi_class'],
      'properties': {
        'solver': {
          'description': 'Algorithm for optimization problem.',
          'enum': ['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'],
          'default': 'liblinear'},
        'penalty': {
          'description': 'Norm used in the penalization.',
          'enum': ['l1', 'l2'],
          'default': 'l2'},
        'dual': {
          'description': 'Dual or primal formulation.',
          'type': 'boolean',
          'default': False},
        'C': {
          'description':
            'Inverse regularization strength. Smaller values specify '
            'stronger regularization.',
          'type': 'number',
          'distribution': 'loguniform',
          'minimum': 0.0,
          'exclusiveMinimum': True,
          'default': 1.0,
          'minimumForOptimizer': 0.03125,
          'maximumForOptimizer': 32768},
        'tol': {
          'description': 'Tolerance for stopping criteria.',
          'type': 'number',
          'distribution': 'loguniform',
          'minimum': 0.0,
          'exclusiveMinimum': True,
          'default': 0.0001,
          'minimumForOptimizer': 1e-05,
          'maximumForOptimizer': 0.1},
        'fit_intercept': {
          'description':
            'Specifies whether a constant (bias or intercept) should be '
            'added to the decision function.',
          'type': 'boolean',
          'default': True},
        'intercept_scaling': {
          'description':
            'Append a constant feature with constant value '
            'intercept_scaling to the instance vector.',
          'type': 'number',
          'distribution': 'uniform',
          'minimum': 0.0,
          'exclusiveMinimum': True,
          'default': 1.0},
        'class_weight': {
          'anyOf': [
            { 'description': 'By default, all classes have weight 1.',
              'enum': [None]},
            { 'description': 'Adjust weights by inverse frequency.',
              'enum': ['balanced']},
            { 'description': 'Dictionary mapping class labels to weights.',
              'type': 'object',
              'propertyNames': {'pattern': '^.+$', 'type': 'number'},
              'forOptimizer': False}],
          'default': None},
        'random_state': {
          'description':
            'Seed of pseudo-random number generator for shuffling data.',
          'anyOf': [
            { 'description': 'RandomState used by np.random',
              'enum': [None]},
            { 'description': 'Explicit seed.',
              'type': 'integer'}],
          'default': None},
        'max_iter': {
          'description':
            'Maximum number of iterations for solvers to converge.',
          'type': 'integer',
          'distribution': 'uniform',
          'minimum': 1,
          'default': 100},
        'multi_class': {
          'description':
            'Approach for more than two classes (not binary classifier).',
          'enum': ['ovr', 'multinomial', 'auto'],
          'default': 'ovr'},
        'verbose': {
          'description':
            'For the liblinear and lbfgs solvers set verbose to any positive '
            'number for verbosity.',
          'type': 'integer',
          'default': 0},
        'warm_start': {
          'description':
            'If true, initialize with solution of previous call to fit.',
          'type': 'boolean',
          'default': False},
        'n_jobs': {
          'description':
            'Number of CPU cores when parallelizing over classes if '
            'multi_class is ovr.',
          'anyOf': [
            { 'description': '1 unless in joblib.parallel_backend context.',
              'enum': [None]},
            { 'description': 'Use all processors.',
              'enum': [-1]},
            { 'description': 'Number of CPU cores.',
              'type': 'integer',
              'minimum': 1}],
            'default': None}}},
      { 'description':
          'The newton-cg, sag, and lbfgs solvers support only l2 penalties.',
        'anyOf': [
          { 'type': 'object',
            'properties': {
              'solver': {'not': {'enum': ['newton-cg', 'sag', 'lbfgs']}}}},
          { 'type': 'object',
            'properties': {'penalty': {'enum': ['l2']}}}]},
      { 'description':
          'The dual formulation is only implemented for l2 '
          'penalty with the liblinear solver.',
        'anyOf': [
          { 'type': 'object',
            'properties': {'dual': {'enum': [False]}}},
          { 'type': 'object',
            'properties': {
              'penalty': {'enum': ['l2']},
              'solver': {'enum': ['liblinear']}}}]},
      { 'description':
          'The multi_class multinomial option is unavailable when the '
          'solver is liblinear.',
        'anyOf': [
          { 'type': 'object',
            'properties': {
              'multi_class': {'not': {'enum': ['multinomial']}}}},
          { 'type': 'object',
            'properties': {
              'solver': {'not': {'enum': ['liblinear']}}}}]}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': """`Logistic regression`_ linear model from scikit-learn for classification.

.. _`Logistic regression`: https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.LogisticRegression.html#sklearn-linear-model-logisticregression
""",
  'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.logistic_regression.html',
  'type': 'object',
  'tags': {
    'pre': ['~categoricals'],
    'op': ['estimator', 'classifier', 'interpretable'],
    'post': ['probabilities']},
  'properties': {
    'hyperparams': _hyperparams_schema,
    'input_fit': _input_fit_schema,
    'input_predict': _input_predict_schema,
    'output_predict': _output_predict_schema,
    'input_predict_proba': _input_predict_proba_schema,
    'output_predict_proba': _output_predict_proba_schema,
    'input_decision_function': _input_decision_function_schema,
    'output_decision_function': _output_decision_function_schema,
}}

class LogisticRegressionImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.linear_model.LogisticRegression(
            **self._hyperparams)

    def fit(self, X, y, **fit_params):
        if fit_params is None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X, y, **fit_params)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)

lale.docstrings.set_docstrings(LogisticRegressionImpl, _combined_schemas)

LogisticRegression = lale.operators.make_operator(LogisticRegressionImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import logging
import numpy
import jsonschema
import os

from lale.util.Visitor import Visitor

from typing import Any, Dict, List, Set, Iterable, Iterator, Optional, Tuple, Union
from lale.schema_simplifier import findRelevantFields, narrowToGivenRelevantFields, simplify, filterForOptimizer

from lale.schema_utils import Schema, getMinimum, getMaximum, STrue, SFalse, is_false_schema, is_true_schema, forOptimizer, has_operator, atomize_schema_enumerations, check_operators_schema
from lale.search.search_space import *
from lale.search.lale_hyperopt import search_space_to_str_for_comparison
from lale.search.PGO import PGO, FrequencyDistribution, Freqs

from lale.operators import *

logger = logging.getLogger(__name__)              

def op_to_search_space(op:PlannedOperator, pgo:Optional[PGO]=None)->SearchSpace:
    """ Given an operator, this method compiles its schemas into a SearchSpace
    """
    search_space = SearchSpaceOperatorVisitor.run(op, pgo=pgo)

    if should_print_search_space("true", "all", "search_space"):
        name = op.name()
        if not name:
            name = "an operator"
        print(f"search space for {name}:\n {str(search_space)}")
    return search_space

def get_default(schema)->Optional[Any]:
    d = schema.get('default', None)
    if d is not None:
        try:
            s = forOptimizer(schema)
            jsonschema.validate(d, s, jsonschema.Draft4Validator)
            return d
        except:
            logger.debug(f"get_default: default {d} not used because it is not valid for the schema {schema}")
            return None
    return None

class FreqsWrapper(object):
    base:Optional[Dict[str,Freqs]]

    def __init__(self, base:Optional[Dict[str,Freqs]]):
        self.base = base

def pgo_lookup(pgo:Optional[PGO], name:str)->Optional[FreqsWrapper]:
    if pgo is None:
        return None
    else:
        freqs:Optional[Dict[str,Freqs]] = None
        if pgo is not None:
            freqs = pgo.get(name, None)
        return FreqsWrapper(freqs)

pgo_part = Union[FreqsWrapper, Freqs, None]

def freqs_wrapper_lookup(part:pgo_part, k:str)->pgo_part:
    if part is None:
        return None
    elif isinstance(part, FreqsWrapper):
        f = part.base
        if f is not None and k in f:
            return f[k]
        else:
            return None
    else:
        return None

def asFreqs(part:pgo_part)->Optional[Iterable[Tuple[Any, int]]]:
    if part is None:
        return None
    elif isinstance(part, FreqsWrapper):
        return None
    else:
        return part.items()

def add_sub_space(space, k, v):
    """ Given a search space and a "key", 
        if the defined subschema does not exist,
        set it to be the constant v space
   """
    # TODO!
    # I should parse __ and such and walk down the schema
    if isinstance(space, SearchSpaceObject):
        if k not in space.keys:
            space.keys.append(k)
            space.choices = (c + (SearchSpaceConstant(v),) for c in space.choices )
            return

# TODO: do we use 'path' above anymore?
# or do we just add the paths later as needed?
class SearchSpaceOperatorVisitor(Visitor):
    pgo:Optional[PGO]

    @classmethod
    def run(cls, op:PlannedOperator, pgo:Optional[PGO]=None)->SearchSpace:
        visitor = cls(pgo=pgo)
        accepting_op:Any = op
        return accepting_op.accept(visitor)

    def __init__(self, pgo:Optional[PGO]=None):
        super(SearchSpaceOperatorVisitor, self).__init__()
        self.pgo = pgo
    
    def visitPlannedIndividualOp(self, op:PlannedIndividualOp)->SearchSpace:
        schema = op._hyperparam_schema_with_hyperparams()
        module = op._impl.__module__
        if module is None or module == str.__class__.__module__:
            long_name = op.name()
        else:
            long_name = module + '.' + op.name()
        name = op.name()
        space = self.schemaToSearchSpace(long_name, name, schema)
        if space is None:
            space = SearchSpaceEmpty()
        # we now augment the search space as needed with the specified hyper-parameters
        # even if they are marked as not relevant to the optimizer, we still want to include them now
        if hasattr(op, '_hyperparams'):
            hyperparams = op._hyperparams
            if hyperparams:
                for (k,v) in hyperparams.items():
                    add_sub_space(space, k, v)
        return space

    visitTrainableIndividualOp = visitPlannedIndividualOp
    visitTrainedIndividualOp = visitPlannedIndividualOp
    
    def visitPlannedPipeline(self, op:'PlannedPipeline')->SearchSpace:
        spaces:List[Tuple[str, SearchSpace]] = [
            (s.name(), s.accept(self)) for s in op.steps()]

        return SearchSpaceProduct(spaces)
    
    visitTrainablePipeline = visitPlannedPipeline
    visitTrainedPipeline = visitPlannedPipeline

    def visitOperatorChoice(self, op:'OperatorChoice')->SearchSpace:
        spaces:List[SearchSpace] = [
            s.accept(self) for s in op.steps()]

        return SearchSpaceSum(spaces)

    # functions to actually convert an individual operator
    # schema into a search space
    def schemaObjToSearchSpaceHelper(self,
        longName:str, 
        path:str, 
        schema:Schema, 
        relevantFields:Optional[Set[str]],
        pgo_freqs:pgo_part=None)->Dict[str,SearchSpace]:
        if 'properties' not in schema:
            return {}
        props = schema['properties']
        hyp:Dict[str, SearchSpace] = {}
        for p,s in props.items():
            if relevantFields is None or p in relevantFields:
                # TODO: This does not handle nested relevant fields correctly
                # We would need to specify what is correct in that case
                sub_freqs = freqs_wrapper_lookup(pgo_freqs, p)
                sub_sch = self.schemaToSearchSpaceHelper_(longName, path + "_" + p, s, relevantFields, pgo_freqs=sub_freqs)
                if sub_sch is None:
                    # if it is a required field, this entire thing should be None
                    hyp[p] = SearchSpaceConstant(None)
                else:
                    hyp[p] = sub_sch
            else:
                logger.debug(f"schemaToSearchSpace: skipping not relevant field {p}")
        return hyp

    def schemaToSearchSpaceHelper_(self,
                                    longName,
                                    path:str, 
                                    schema:Schema, 
                                    relevantFields:Optional[Set[str]],
                                    pgo_freqs:pgo_part=None)->Optional[SearchSpace]:
        # TODO: handle degenerate cases
        # right now, this handles only a very fixed form

        if is_false_schema(schema):
            return None

        typ:Optional[str] = None
        typ = schema.get('laleType', None)
        if typ is None:
            typ = schema.get('type', None)
        else:
            typ = typ

        if 'enum' in schema and typ != "operator":
            vals = schema['enum']
            return SearchSpaceEnum(vals, pgo=asFreqs(pgo_freqs), default=get_default(schema))

        if typ is not None:
            if typ == "boolean":
                return SearchSpaceBool(pgo=asFreqs(pgo_freqs), default=get_default(schema))
            elif typ == "number" or typ == "integer":
                exclusive_minimum = False
                minimum=schema.get('minimumForOptimizer', None)
                if minimum is not None:
                    exclusive_minimum = schema.get('exclusiveMinimumForOptimizer', False)
                else:
                    minimum=schema.get('minimum', None)
                    if minimum is not None:
                        exclusive_minimum = schema.get('exclusiveMinimum', False)

                exclusive_maximum = False
                maximum=schema.get('maximumForOptimizer', None)
                if maximum is not None:
                    exclusive_maximum = schema.get('exclusiveMaximumForOptimizer', False)
                else:
                    maximum=schema.get('maximum', None)
                    if maximum is not None:
                        exclusive_maximum = schema.get('exclusiveMaximum', False)

                distribution = schema.get('distribution', None)

                laleType = schema.get('laleType', None)
                if laleType is None:
                    laleType = typ

                if laleType == "number":
                    discrete = False
                elif laleType == "integer":
                    discrete = True
                else:
                    raise NotImplementedError()
                
                pgo:Freqs

                return SearchSpaceNumber(minimum=minimum, 
                                exclusiveMinimum=exclusive_minimum, 
                                maximum=maximum, 
                                exclusiveMaximum=exclusive_maximum, 
                                discrete=discrete, 
                                distribution=distribution,
                                pgo=asFreqs(pgo_freqs),
                                default=get_default(schema))
            elif typ == "array" or typ =="tuple":
                laleType = schema.get('laleType', None)
                if laleType is None:
                    laleType = typ

                is_tuple:bool = laleType == "tuple"

                min_items = schema.get('minItemsForOptimizer', None)
                if min_items is None:
                    min_items = schema.get('minItems', None)
                    if min_items is None:
                        min_items = 0
                max_items = schema.get('maxItemsForOptimizer', None)
                if max_items is None:
                    max_items = schema.get('maxItems', None)

                items_schema = schema.get('itemsForOptimizer', None)
                if items_schema is None:
                    items_schema = schema.get('items', None)
                    if items_schema is None:
                        raise ValueError(f"an array type was found without a provided schema for the items in the schema {schema}.  Please provide a schema for the items (consider using itemsForOptimizer)")

                # we can search an empty list even without schemas
                if max_items == 0:
                    if is_tuple:
                        return SearchSpaceConstant([()])
                    else:
                        return SearchSpaceConstant([[]])

                prefix:Optional[List[SearchSpace]] = None
                additional:Optional[SearchSpace] = None
                if isinstance(items_schema, list):
                    prefix = []
                    for i,sub_schema in enumerate(items_schema):
                        sub = self.schemaToSearchSpaceHelper_(longName, path + "_" + str(i), sub_schema, relevantFields)
                        if sub is None:
                            return None
                        else:
                            prefix.append(sub)
                    prefix_len = len(prefix)
                    additional_items_schema = schema.get('additionalItemsForOptimizer', None)
                    if additional_items_schema is None:
                        additional_items_schema = schema.get('additionalItems', None)
                    if additional_items_schema is None:
                        if max_items is None or max_items > prefix_len:
                            raise ValueError(f"an array type was found with provided schemas for {prefix_len} elements, but either an unspecified or too high a maxItems, and no schema for the additionalItems.  Please constraing maxItems to <= {prefix_len} (you can set maxItemsForOptimizer), or provide a schema for additionalItems")
                    elif additional_items_schema == False:
                        if max_items is None:
                            max_items = prefix_len
                        else:
                            max_items = min(max_items, prefix_len)
                    else:
                        additional = self.schemaToSearchSpaceHelper_(longName, path + "-", sub_schema, relevantFields)
                        # if items_schema is None:
                        #     raise ValueError(f"an array type was found without a provided schema for the items in the schema {schema}.  Please provide a schema for the items (consider using itemsForOptimizer)")
                else:
                    additional = self.schemaToSearchSpaceHelper_(longName, path + "-", items_schema, relevantFields)

                if max_items is None:
                    raise ValueError(f"an array type was found without a provided maximum number of items in the schema {schema}, and it is not a list with 'additionalItems' set to False.  Please provide a maximum (consider using maxItemsForOptimizer), or, if you are using a list, set additionalItems to False")

                return SearchSpaceArray(prefix=prefix, minimum=min_items, maximum=max_items, additional=additional, is_tuple=is_tuple)

            elif typ == "object":
                if 'properties' not in schema:
                    return SearchSpaceObject(longName, [], [])
                o = self.schemaObjToSearchSpaceHelper(longName, path, schema, relevantFields, pgo_freqs=pgo_freqs)
                all_keys = list(o.keys())
                all_keys.sort()
                o_choice = tuple([o.get(k, None) for k in all_keys])
                return SearchSpaceObject(longName, all_keys, [o_choice])
            elif typ == "string":
                pass
            elif typ == "operator":
                # TODO: If there is a default, we could use it
                vals = schema.get('enum', None)
                if vals is None:
                    logger.error(f"An operator is required by the schema but was not provided")
                    return None
                
                sub_schemas = [op.accept(self) for op in vals]
                combined_sub_schema:SearchSpace
                if len(sub_schemas) == 1:
                    combined_sub_schema = sub_schemas[0]
                else:
                    combined_sub_schema = SearchSpaceSum(sub_schemas)
                return SearchSpaceOperator(combined_sub_schema)

            elif typ == "Any":
                raise ValueError(f"In path: {path}, longName: {longName}: A search space was found with laleType ({typ}), which is not searchable.  Please mark the relevant hyperparameter as not relevant for the optimizer.  schema: {schema}")
            else:
                raise ValueError(f"An unknown type ({typ}) was found in the schema {schema}")

        if 'anyOf' in schema:
            objs = []
            for s_obj in schema['anyOf']:
                if 'type' in s_obj and s_obj['type'] == "object":
                    o = self.schemaObjToSearchSpaceHelper(longName, path, s_obj, relevantFields, pgo_freqs=pgo_freqs)
                    if o: objs.append(o)
            if objs:
                # First, gather a list of all the properties
                keys_list = [set(o.keys()) for o in objs]
                # make sure the iterator is deterministic
                all_keys = list(set.union(*keys_list))
                # and we might as well make it sorted
                all_keys.sort()
                def as_str(k, c):
                    if c is None:
                        return "None"
                    else:
                        return search_space_to_str_for_comparison(c, path + "_" + k)
                anys:Dict[str,Any] = {}
                for o in objs:
                    o_choice = tuple([o.get(k, None) for k in all_keys])
                    k = str([as_str(all_keys[idx], c) for idx, c in enumerate(o_choice)])
                    if k in anys:
                        logger.info(f"Ignoring Duplicate SearchSpace entry {k}")
                    anys[k] = o_choice
                return SearchSpaceObject(longName, all_keys, anys.values())
            else:
                return SearchSpaceObject(longName, [], [])
        
        if 'allOf' in schema:
            # if all but one are negated constraints, we will just ignore them
            pos_sub_schema:List[Schema] = []
            for sub_schema in schema['allOf']:
                if 'not' not in sub_schema:
                    pos_sub_schema.append(sub_schema)

            if len(pos_sub_schema) > 1:
                raise ValueError(f"schemaToSearchSpaceHelper does not yet know how to compile the given schema {schema} for {longName}, because it is an allOf with more than one non-negated schemas ({pos_sub_schema})")
            if len(pos_sub_schema) == 0:
                raise ValueError(f"schemaToSearchSpaceHelper does not yet know how to compile the given schema {schema} for {longName}, because it is an allOf with only negated schemas")
            
            logger.debug(f"schemaToSearchSpaceHelper: ignoring negated schemas in the conjunction {schema} for {longName}")
            return self.schemaToSearchSpaceHelper_(longName, 
                                    path,
                                    pos_sub_schema[0], 
                                    relevantFields,
                                    pgo_freqs=pgo_freqs)
        # TODO: handle degenerate cases
        raise ValueError(f"schemaToSearchSpaceHelper does not yet know how to compile the given schema {schema} for {longName}")

    def schemaToSearchSpaceHelper(self,
                                longName, 
                                schema:Schema, 
                                relevantFields:Optional[Set[str]],
                                pgo_freqs:pgo_part=None)->Optional[SearchSpace]:
        if not is_false_schema(schema) and not schema:
            return None
        else:
            return self.schemaToSearchSpaceHelper_(longName, longName, schema, relevantFields, pgo_freqs=pgo_freqs)


    def schemaToSimplifiedAndSearchSpace(self,
        longName:str, 
        name:str, 
        schema:Schema)->Tuple[Schema, Optional[SearchSpace]]:
        relevantFields = findRelevantFields(schema)
        if relevantFields:
            schema = narrowToGivenRelevantFields(schema, relevantFields)

        if has_operator(schema):
            atomize_schema_enumerations(schema)
        simplified_schema = simplify(schema, True)
    #    from . import pretty_print
    #    print(f'SIMPLIFIED_{longName}: {pretty_print.to_string(simplified_schema)}')

        filtered_schema = filterForOptimizer(simplified_schema)
    #    print(f'SIMPLIFIED_{longName}: {pretty_print.to_string(filtered_schema)}')

        if logger.isEnabledFor(logging.WARNING):
            op_warnings:List[str] = []
            check_operators_schema(filtered_schema, op_warnings)
            if op_warnings:
                for w in op_warnings:
                    logger.warning(w)

        return (filtered_schema, 
                self.schemaToSearchSpaceHelper(
                            longName, 
                            filtered_schema, 
                            relevantFields, 
                            pgo_freqs=pgo_lookup(self.pgo, name)))

    def schemaToSearchSpace(self,
                            longName:str, 
                            name:str, 
                            schema:Schema
                            )->Optional[SearchSpace]:
        (s, h) = self.schemaToSimplifiedAndSearchSpace(longName, name, schema)
        return h
